"""
The methods to run the parts of the form compiler

Should also contain the entrypoint methods.
"""
from __future__ import absolute_import

import loopy

from ufl.algorithms import compute_form_data, read_ufl_file
from ufl.algorithms.formfiles import interpret_ufl_namespace

from dune.perftool.generation import delete_cache_items, global_context
from dune.perftool.interactive import start_interactive_session
from dune.perftool.options import get_option
from dune.perftool.pdelab.driver import generate_driver
from dune.perftool.pdelab.localoperator import (generate_localoperator_basefile,
                                                generate_localoperator_file,
                                                generate_localoperator_kernels,
                                                name_localoperator_file)
from dune.perftool.ufl.preprocess import preprocess_form

from os.path import splitext, basename


# Disable loopy caching before we do anything else!
loopy.CACHING_ENABLED = False


def type_guessing(val):
    for t in [int, float]:
        try:
            return t(val)
        except ValueError:
            pass
    return val


def read_ufl(uflfile):
    """Read uflfile file, extract and preprocess forms

    Arguments:
    ----------
    uflfile: Path to ufl file.

    Returns:
    --------
    formdatas: List of formdatas found in uflfile.
    forms: List of forms found in uflfile.
    """
    # Read the given ufl file and execute it
    uflcode = read_ufl_file(uflfile)

    # Prepopulate a namespace with variation information
    namespace = globals()
    ini = get_option("ini_file")
    if ini:
        from dune.common.parametertree.parser import parse_ini_file
        ini = parse_ini_file(ini)

        for k, v in ini.get("formcompiler.ufl_variants", {}).items():
            namespace[k] = type_guessing(v)

    try:
        exec("from dune.perftool.ufl.execution import *\n" + uflcode, namespace)
    except:
        name = splitext(basename(uflfile))[0]
        name = "{}_debug".format(name)
        pyname = "{}.py".format(name)
        print(pyname)
        pycode = "#!/usr/bin/env python\nfrom dune.perftool.ufl.execution import *\nset_level(DEBUG)\n"
        for k, v in ini.get("formcompiler.ufl_variants", {}).items():
            pycode = pycode + "{} = {}\n".format(k, repr(type_guessing(v)))
        pycode = pycode + uflcode
        with file(pyname, "w") as f:
            f.write(pycode)
        raise SyntaxError("Not a valid ufl file, dumped a debug script: {}".format(pyname))

    # Extract and preprocess the forms
    data = interpret_ufl_namespace(namespace)

    # Enrich data by some additional objects we deem worth keeping
    if get_option("exact_solution_expression"):
        data.object_by_name[get_option("exact_solution_expression")] = namespace[get_option("exact_solution_expression")]

    formdatas = []
    forms = data.forms
    for index, form in enumerate(forms):
        formdatas.append(preprocess_form(form))
        forms[index] = formdatas[index].preprocessed_form

    # We expect at least one form
    assert len(data.forms) >= 1

    return formdatas, data


# This function is the entrypoint of the ufl2pdelab executable
def compile_form():
    formdatas, data = read_ufl(get_option("uflfile"))

    with global_context(data=data, formdatas=formdatas):
        # Generate driver file
        if get_option("driver_file"):
            generate_driver(formdatas, data)

    # In case of multiple forms: Genarate one file that includes all localoperator files
    if len(formdatas) > 1:
        generate_localoperator_basefile(formdatas, data)

    # Generate local operator files
    for formdata in formdatas:
        with global_context(data=data, formdata=formdata):
            # Make sure cache is empty
            delete_cache_items()

            # Create localoperator kernels
            if get_option("operator_file"):
                kernels = generate_localoperator_kernels(formdata, data)

            # TODO insert sophisticated analysis/feedback loops here
            if get_option("interactive"):
                start_interactive_session(kernels)

            # Create c++ file from kernels
            if get_option("operator_file"):
                filename = name_localoperator_file(formdata, data)
                generate_localoperator_file(formdata, kernels, filename)
